package edu.uci.ics.pattis.ics23.collections;


import java.util.ConcurrentModificationException;
import java.util.Iterator;
import java.util.NoSuchElementException;



//In this Abstract Class, many concrete methods can be written by using
//an iterator; generally these methods are slow and should be
//overridden in any concrete classes that extend this one.

public abstract class AbstractMap<K,V> implements Map<K,V> {

	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public void putAll (Iterable<Map.Entry<K,V>> es) {
	  for (Entry<? extends K,? extends V> e : es)
	    put(e.getKey(),e.getValue());
  }

	
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public V remove(K k) {
		Iterator<Entry<K,V>> i = entries().iterator();
		Entry<K,V> e;
		if (k == null)
			while (i.hasNext()) {
        e = i.next();
				if (e.getKey() == null) {
					i.remove();
					return e.getValue();
				}
			}
		else
			while (i.hasNext()) {
				e = i.next();
				if (k.equals(e.getKey())) {
					i.remove();
					return e.getValue();
			}
		}

		return null;
	}

  
	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public void clear () {
    Iterator<Entry<K,V>> i = entries().iterator();
    while (i.hasNext()) {
    	i.next();
    	i.remove();
    }
  }
  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public V get(Object key) {
  	if (key == null) {
			for (Entry<K,V> e : entries())
				if (e.getKey()==null)
					return e.getValue();
		}else
			for (Entry<K,V> e : entries()) {
				if (key.equals(e.getKey()))
					return e.getValue();
			}
    return null;
  }   
  
  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public boolean containsKey(Object key) {
		if (key == null) {
			for (Entry<K,V> e : entries())
				if (e.getKey() == null)
					return true;
		}else
			for (Entry<K,V> e : entries())
				if (key.equals(e.getValue()))
					return true;

		return false;
	}

	
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
	public boolean containsValue(Object value) {
		if (value == null) {
			for (Entry<K,V> e : entries())
				if (e.getValue() == null)
					return true;
		}else
			for (Entry<K,V> e : entries())
				if (value.equals(e.getValue()))
					return true;

		return false;
	}
	

	
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  public Iterable<K> keys () {
    return new Iterable<K>() {
      public Iterator<K> iterator() {
        return new Iterator<K>() {
          private Iterator<Map.Entry<K,V>> base = entries().iterator();
          public boolean hasNext()
          {return base.hasNext();}
          
          public K next() throws ConcurrentModificationException,NoSuchElementException
          {return base.next().getKey();}

          public void remove() throws UnsupportedOperationException, IllegalStateException
          {base.remove();}
      };
      }
    };
  }

  
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
 public Iterable<V> values () {
   return new Iterable<V>() {
     public Iterator<V> iterator() {
       return new Iterator<V>() {
         private Iterator<Map.Entry<K,V>> base = entries().iterator();
         public boolean hasNext()
         {return base.hasNext();}
         
         public V next() throws ConcurrentModificationException,NoSuchElementException
         {return base.next().getValue();}

         public void remove() throws UnsupportedOperationException, IllegalStateException
          {base.remove();}
     };
     }
   };
 }
 
 
 
  //Generally, inherit this method without overriding
  //Ensure that size() computes its result quickly (see below)
  public boolean isEmpty()
  {return size() == 0;}

	
  
  //Override this method if there is a more efficient algorithm
  //  for a concrete data structure
  //It is OK that e is never used in the for loop
  @SuppressWarnings("unused")
  public int size() {
    int count = 0;
    for (K e : keys())
    	count++;
    return count;
  }


	
  //Generally, inherit this method without overriding
  //It is OK that the new constructed Map.Entry<?,?>[] is cast to
  //  Map.Entry<K,V>[] because it is filled with Map.Entry<K,V> objects
  @SuppressWarnings("unchecked")
  public Map.Entry<K,V>[] toArray() {
  	Map.Entry<K,V>[] result = (Map.Entry<K,V>[]) new Map.Entry<?,?>[size()];
  	int i = 0;
	  for (Map.Entry<K,V> e : entries())
	    result[i++] = e;
	  return result;
	}

  

  //Override this method to add just the name of the concrete class
  //  and/or to illustrate the data structure used in the implementation
  //  without relying on an iterator for that class
  public String toString() {
   	StringBuffer answer = new StringBuffer();
     for (Iterator<Map.Entry<K,V>> i = entries().iterator(); i.hasNext(); /*See i.next() in body*/) {
       Map.Entry<K,V> e = i.next();
       answer.append( (e == this ? "(this Map)" : String.valueOf(e)) +
       		           (i.hasNext() ? "," : "")
                     );
     }
  	  return answer.toString();
 	}
 
  
  
  //Generally, inherit this method without overriding
  public boolean equals(Object o) {
  	//Checks external to values in collection
  	if (o == this)
  		return true;
  	if (!(o instanceof Map<?,?>))
  		return false;
   	Map<?,?> m = (Map<?,?>) o;
  	if (m.size() != size())
  		return false;

		//Checks internal to values in collection: both the same size
   	try {
  		for (Entry<K,V> e : entries()) {
  			K key   = e.getKey();
  			V value = e.getValue();
  			if (value == null) {
  				if (!(m.get(key)==null && m.containsKey(key)))
  					return false;
  			} else {
  				if (!value.equals(m.get(key)))
  					return false;
  			}
  		}
  	} catch (ClassCastException unused) {
  		return false;
  	} catch (NullPointerException unused) {
  		return false;
  	}

  	return true;
  }

 
  //Order is NOT important
  public int hashCode() {
  	int answer = 0;
  	for (Entry<K,V> e : entries())
  		answer += (e==null ? 0 : e.hashCode());
  	return answer;
  }
 

  
  //These methods must be available in all concrete classes
	public abstract V                        put         (K key, V value);
  public abstract Iterable<Map.Entry<K,V>> entries     ();
  public abstract Map<K,V>                 newEmpty    ();
  public abstract Map<K,V>                 shallowCopy ();
 
  
  
  //Used in subclasses: see their iterators
  protected transient volatile long modCount = 0;
 
  
  
  protected static class SimpleEntry<K,V> implements Entry<K,V> {
  	private final K key;
  	private       V value;
  	public SimpleEntry(K key, V value) {
  		this.key   = key;
  		this.value = value;
  	}

  	public SimpleEntry(Entry<? extends K, ? extends V> entry) {
  		this.key   = entry.getKey();
  		this.value = entry.getValue();
  	}

  	public K getKey()
  	{return key;}

  	public V getValue()
  	{return value;}

  	public V setValue(V newValue) {
  		V oldValue = value;
  		value = newValue;
  		return oldValue;
  	}

  	public boolean equals(Object o) {
  		if (!(o instanceof Map.Entry<?,?>))
  			return false;
  		Map.Entry<?,?> e = (Map.Entry<?,?>)o;
  		return (key   == null ? e.getKey()   == null : key  .equals(e.getKey())) &&
  				   (value == null ? e.getValue() == null : value.equals(e.getValue()));
  	}

  	public int hashCode() {
  		return (key   == null ? 0 :   key.hashCode()) ^
  		(value == null ? 0 : value.hashCode());
  	}

  	public String toString() {
  		return key + "->" + value;
  	}

  }

}
