package edu.uci.ics.pattis.ics23.collections;


import java.io.File;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import java.util.Arrays;

import edu.uci.ics.pattis.introlib.Prompt;




public class DriverMagic {

  private static boolean classImplementsInterface(Class<?> aClass, String anInterface) {
     if (aClass == null)
      return false;
  
     if (aClass.toString().endsWith("."+anInterface))
       return true;
      for (Class<?> c : aClass.getInterfaces()) {
        if (c.toString().endsWith("."+anInterface))
            return true;
        for (Class<?> c2 : c.getInterfaces())
          if (classImplementsInterface(c2,anInterface))
            return true;
      }
    
    return false;
  }

  

  public static String[] getClasses(String packageName, String anInterface)
  throws ClassNotFoundException {

    Set<Class<?>> classes = new ArraySet<Class<?>>();

    //Find the directories/jar files that contain all classes in a certain package
    File directory;
    try {
      Enumeration<URL> rs = Thread.currentThread().getContextClassLoader().getResources(packageName.replace('.', '/'));
      while (rs.hasMoreElements()) {
        String dirName = rs.nextElement().getFile().replace("%20"," ");

        //Jar file or Directory?
        if (dirName.contains(".jar!")) {
          dirName = dirName.substring(5,dirName.indexOf(".jar!")+4);
          //In this Jar file (if it exists), examine all files ending in .class, and
          //  add concrete (non-interface, non-abstract) classes that implement
          //  anInterface
          try { 
            ZipFile zip = new ZipFile(new File(dirName));
            for (Enumeration<? extends ZipEntry> e = zip.entries(); e.hasMoreElements();) { 
              String entry = ((ZipEntry) e.nextElement()).getName().replace('/','.'); 
              if(entry.endsWith(".class") && entry.indexOf('$') == -1) {
                Class<?> c = Class.forName(entry.substring(0, entry.length()-6));
                if (0 == (c.getModifiers() & (Modifier.ABSTRACT | Modifier.INTERFACE)) &&
                    classImplementsInterface(c,anInterface))
                  classes.add(c);
              }
            } 
          }
          catch (ZipException e) 
          {e.printStackTrace();}
          catch (IOException e) 
          {e.printStackTrace();} 
          
        }else{
          
          //In this directory (if it exists), examine all files ending in .class, and
          //  add concrete (non-interface, non-abstract) classes that implement
          //  anInterface
          directory = new File(dirName);
          if(directory.exists()) {
            String[] files = directory.list();
            for(int i=0; i<files.length; i++){
              if(files[i].endsWith(".class") && files[i].indexOf('$') == -1) {
                Class<?> c = null;
                try {
                  c = Class.forName(packageName+'.'+files[i].substring(0, files[i].length()-6));
                  if (0 == (c.getModifiers() & (Modifier.ABSTRACT | Modifier.INTERFACE)) &&
                     classImplementsInterface(c,anInterface))
                    classes.add(c);
                } catch (Exception e)
                {System.out.println("Failed on class = " + c +": "+e.getMessage());}
              }
            }
          } else {
            throw new ClassNotFoundException(packageName+" does not appear to be a valid package");
          }

        }
      }
    } catch(IOException e)
      {e.printStackTrace();}
      catch(NullPointerException e) {
      throw new ClassNotFoundException(packageName+" does not appear to be a valid package");
    }


    //Convert classes to an array of String, keeping only the class name
    // (losing the package name)
    String[] returnAnswer = new String[classes.size()];
    int i = 0;
    for (Class<?> c : classes)
      returnAnswer[i++] = c.toString().substring(c.toString().lastIndexOf(".")+1);
    Arrays.sort(returnAnswer);
    return returnAnswer;
  }  

  
  
  public static Class<?> getClass(String packageName, String kind, String[] classes) {
    int tries = 1;
    for (;;)
      try {
        String className;
        if (tries >1 || classes.length != 1)
          className = Prompt.forStringViaIndex("Enter \""+kind+"\" Class to Test",classes);
        else {
          className = classes[0];
          System.out.println("Trying to test " + className + ", the only class found implementing the correct interface");
        }

        tries++;
        Class<?> c = Class.forName(packageName+"."+className);
        return (Class<?>)c;
        
      }catch(Exception e) {
        System.out.println("  Cannot locate this class");
    }
  }
  
}
